(* W01 *)
(* E01 *)
(* W02 *) 
type 'a t = 'a list
(* hd : 'a list -> 'a *)
let hd = function x::_ -> x
                | _    -> failwith "empty list"
(* tl : 'a list -> 'a list *)
let tl = function _::xs -> xs
                | _     -> failwith "empty list"
(* append : 'a list -> 'a list -> 'a list *)
let rec (@) xs ys = match xs with []     -> ys
                                 | x::xs -> x::(xs @ ys)
(* replicate : int -> 'a -> 'a list *)
let rec replicate n x = 
 if n < 1 then [] else x::replicate (n-1) x
(* take : int -> 'a list -> 'a list *)
let rec take n xs = if n < 1 then [] else match xs with
 | []    -> []
 | x::xs -> x::take (n-1) xs
(* drop : int -> 'a list -> 'a list *)
let rec drop n xs = if n < 1 then xs else match xs with
 | []    -> []
 | _::xs -> drop (n-1) xs
(* split_at : int -> 'a list -> ('a list * 'a list *)
let split_at n xs = (take n xs,drop n xs)
(* map : ('a -> 'b) -> 'a list -> 'b list *)
let rec map f = function []    -> []
                       | x::xs -> f x::map f xs
(* foldr : ('a -> 'b -> 'b) -> 'b -> 'a list -> 'b *)
let rec foldr f b = function []     -> b
                           | x::xs -> f x (foldr f b xs)
(* foldl : ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a *)
let rec foldl f b = function []    -> b
                           | x::xs -> foldl f (f b x) xs
(* filter : ('a -> bool) -> 'a list -> 'a list *)
let rec filter p = function
 | []    -> []
 | x::xs -> if p x then x::filter p xs else filter p xs
(* E02 *)
(* mem : 'a -> 'a list -> bool *)
let rec mem x = function []    -> false
                       | y::ys -> y = x || mem x ys
(* length : 'a list -> int *)
let rec length = function []    -> 0
                        | _::xs -> 1 + length xs
(* last : 'a list -> 'a *)
let rec last = function
 | []    -> failwith "empty list"
 | [x]   -> x
 | _::xs -> last xs
(* nth : 'a list -> int -> 'a *)
let rec nth xs i =
 if i < 0 then failwith "out of bounds" else match xs with
  | []    -> failwith "out of bounds"
  | x::xs -> if i = 0 then x else nth xs (i-1)
(* reverse : 'a list ->  'a list *)
let rec reverse = function    [] -> []
                         | x::xs -> (reverse xs) @ [x]
(* rev_append : 'a list -> 'a list -> 'a list *)
let rec rev_append xs ys = match xs with []    -> ys
                                       | x::xs -> rev_append xs (x::ys)
(* concat : 'a list list -> 'a list *)
let concat xs = foldr (@) [] xs
(* zip_with : ('a -> 'b -> c) -> 'a list -> 'b list -> 'c list *)
let rec zip_with f xs ys = match (xs,ys) with
 | (x::xs,y::ys) -> f x y :: zip_with f xs ys
 | _             -> []
(* zip : 'a list -> 'b list -> ('a * 'b)list *)
let zip xs ys = zip_with (fun x y -> (x,y)) xs ys
(* foldl1 : ('a -> 'a -> 'a) -> 'a list -> 'a *)
let foldl1 f = function []    -> failwith "empty list"
                      | x::xs -> foldl f x xs
(* foldr1 : ('a -> 'a -> 'a) -> 'a list -> 'a *)
let rec foldr1 f = function []    -> failwith "empty list"
                          | [x]   -> x
                          | x::xs -> f x (foldr1 f xs)
(* W03 *) 
(* join : 'a list -> 'a list list -> 'a list *)
let join delim = foldr1(fun xs ys -> xs@delim@ys)
(* E03 *) 
(* W04 *)
(* take_while : ('a -> bool) -> 'a list -> 'a list *)
let rec take_while p = function
 | []    -> []
 | x::xs -> if p x then x :: take_while p xs else []
(* drop_while : ('a -> bool) -> 'a list -> 'a list *)
let rec drop_while p = function
 | []            -> []
 | x::xs as list -> if p x then drop_while p xs else list
(* span : ('a -> bool) -> 'a list -> ('a list,'a list) *)
let span p xs = (take_while p xs, drop_while p xs)
(* rev : 'a list -> 'a list *)
let rev xs =
 let rec rev acc = function []    -> acc
                          | x::xs -> rev (x::acc) xs
 in
 rev [] xs
(* until : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a *)
let rec until p f x = if p x then x else until p f (f x)
(* E04 *) 
(* zip4_with : ('a -> 'b -> 'c -> 'd -> 'e) -> 'a t -> 'b t -> 'c t -> 'd t -> 'e t *)
let rec zip4_with f ws xs ys zs = match (ws,xs,ys,zs) with
 | (w::ws,x::xs,y::ys,z::zs) -> f w x y z :: zip4_with f ws xs ys zs
 | _                         -> []
(* zip4 : 'a t -> 'b t -> 'c t -> 'e t -> ('a * 'b * 'c * 'd)t *)
let zip4 ws xs ys zs = zip4_with (fun w x y z -> (w,x,y,z)) ws xs ys zs
(* W05 *)
(* E05 *)
(* W06 *)
(* E06 *)
(* W07 *)
(* E07 *)
(* W08 *)
(* E08 *)
(* W09 *)
